@@ -1,5 +1,5 @@
+import typing
 from pathlib import Path
-from typing import Optional
 
 import typer
 
@@ -66,13 +66,13 @@ def start(
         ),
     ),
     shell: str = typer.Option(
-        '/usr/bin/sh -l',
+        None,
         '--shell',
         '-s',
         show_default=True,
         help='Which remote shell binary to use.',
     ),
-    version: Optional[bool] = typer.Option(
+    version: typing.Optional[bool] = typer.Option(
         None,
         '--version',
         callback=version_callback,
